// Reporting successor states -*- c++ -*-

/** @file StateReporter.C
 * Interface for reporting successor states
 */

#ifdef __GNUC__
# pragma implementation
#endif // __GNUC__

#include "StateReporter.h"
#include "States.h"
#include "Graph.h"
#include "Net.h"
#include "Transition.h"
#include "Valuation.h"
#include "GlobalMarking.h"
#include "FullSet.h"
#ifdef EXPR_COMPILE
# include "Compilation.h"
# include "Printer.h"
#endif // EXPR_COMPILE

/** Flag: has the analysis been interrupted? */
volatile bool interrupted = false;

StateReporter::StateReporter (
#ifdef EXPR_COMPILE
			      const class Compilation* compilation,
#endif // EXPR_COMPILE
			      const class Net& net_,
			      const class Printer& printer_,
			      bool compress,
			      bool local) :
#ifdef EXPR_COMPILE
  myCompilation (compilation),
#endif // EXPR_COMPILE
  net (net_), printer (printer_),
  myEnabled (net.getNumEnabled ()
	     ? new char[net.getNumEnabled ()]
	     : 0),
  myFatal (false), mySrcDeadlock (true),
  myPriority (0), myStateBuf (),
  myLocal (local ? new class FullSet () : 0),
  mySuppressed (compress ? new class States () : 0),
  mySuccessors (None), myMinSize (size_t (-1L)), myMaxSize (0)
{
  if (net.getNumEnabled ())
    memset (myEnabled, 0, net.getNumEnabled ());
  if (myLocal && !myLocal->init ()) {
    delete myLocal;
    const_cast<class FullSet*&>(myLocal) = 0;
  }
}

StateReporter::StateReporter (const class StateReporter& old) :
#ifdef EXPR_COMPILE
  myCompilation (old.myCompilation),
#endif // EXPR_COMPILE
  net (old.net), printer (old.printer),
  myEnabled (net.getNumEnabled ()
	     ? new char[net.getNumEnabled ()]
	     : 0),
  myFatal (false), mySrcDeadlock (true),
  myPriority (0), myStateBuf (),
  myLocal (old.myLocal ? new class FullSet () : 0),
  mySuppressed (old.mySuppressed ? new class States () : 0),
  mySuccessors (None), myMinSize (size_t (-1L)), myMaxSize (0)
{
  if (net.getNumEnabled ())
    memset (myEnabled, 0, net.getNumEnabled ());
  if (myLocal && !myLocal->init ()) {
    delete myLocal;
    const_cast<class FullSet*&>(myLocal) = 0;
  }
}

StateReporter::~StateReporter ()
{
  delete[] myEnabled;
  delete mySuppressed;
  delete myLocal;
}

#ifdef EXPR_COMPILE
/** Report an encoded state
 * @param buf	the encoded state
 * @param size	length of the encoded state, in bytes
 * @param err	errNone if the state is not rejected
 * @param tr	the number of the transition that fired
 * @param hide	flag: is the transition hidden?
 * @param ctx	the call-back context
 */
static void
addState (const void* buf, size_t size, enum Error err,
	  unsigned tr, int hide, void* ctx)
{
  class StateReporter& reporter = *static_cast<class StateReporter*>(ctx);
  if (err && err != errComp)
    reporter.flagFatal ();
  reporter.addState (buf, size, err != errNone, hide, tr);
}

void
StateReporter::addState (const void* buf, size_t size,
			 bool rejected, bool hidden, unsigned tr)
{
  const class Transition& t = net.getTransition (tr);
  assert (!myPriority || t.getPriority () == myPriority);
  myPriority = t.getPriority ();
  mySrcDeadlock = false;
  report (buf, size, rejected, hidden);
}
#endif // EXPR_COMPILE

void
StateReporter::analyze (enum SearchKind kind)
{
  const unsigned numTr = net.getNumTransitions ();
#ifdef EXPR_COMPILE
  if (myCompilation) {
    myCompilation->linkAddState (&::addState, isGraph ());
    myCompilation->getFatal () = &myFatal;
    while (!interrupted && !myFatal) {
      if (!popCompiled (kind == Breadth))
	break;
      mySrcDeadlock = true, myPriority = 0;
      for (unsigned i = 0; !interrupted && !myFatal && i < numTr; i++) {
	const class Transition& t = net.getTransition (i);
	if (!myPriority || t.getPriority () == myPriority) {
	  if (unsigned errors =
	      myCompilation->eventAnalyze (i, this, myEnabled)) {
	    reject ();
	    printer.printQuoted (t.getName ());
	    printer.delimiter (':');
	    printer.print (errors);
	    printer.printRaw (errors == 1 ? " error" : " errors");
	    printer.finish ();
	  }
	}
      }
      if (mySrcDeadlock && !interrupted && !myFatal) {
	switch (myCompilation->stateDeadlock ()) {
	case errNone:
	  break;
	default:
	  myFatal = true;
	  // fall through
	case errComp:
	  reportDeadlock ();
	}
      }
      if (myLocal && myLocal->getNumStates ())
	continue;
      if (mySuppressed) {
	switch (mySuccessors) {
	case One:
	  continue;
	case Many:
	  if (const size_t num = mySuppressed->size ()) {
	    const struct States::state& s = (*mySuppressed)[num - 1];
	    do_addState (s.buf, s.size);
	    mySuccessors = None;
	  }
	  // fall through
	case None:
	  mySuppressed->clear ();
	}
      }
      addEvents ();
      if (kind >= Single)
	break;
    }
    return;
  }
#endif // EXPR_COMPILE
  while (!interrupted && !myFatal) {
    class GlobalMarking* m = pop (kind == Breadth);
    if (!m)
      break;
    mySrcDeadlock = true, myPriority = 0;
    for (unsigned i = 0; !interrupted && !myFatal && i < numTr; i++) {
      const class Transition& t = net.getTransition (i);
      if (!myPriority || t.getPriority () == myPriority)
	t.analyze (*m, *this);
    }
    if (mySrcDeadlock && !interrupted && !myFatal) {
      switch (net.isDeadlock (*m)) {
      case Net::OK:
	break;
      case Net::Fatal:
	myFatal = true;
	// fall through
      case Net::Error:
	reportDeadlock ();
      }
    }
    delete m;
    if (myLocal && myLocal->getNumStates ())
      continue;
    if (mySuppressed) {
      switch (mySuccessors) {
      case One:
	continue;
      case Many:
	if (const size_t num = mySuppressed->size ()) {
	  const struct States::state& s = (*mySuppressed)[num - 1];
	  do_addState (s.buf, s.size);
	  mySuccessors = None;
	}
	// fall through
      case None:
	mySuppressed->clear ();
      }
    }
    addEvents ();
    if (kind >= Single)
      break;
  }
}

bool
StateReporter::report (const class Transition& transition,
		       const class Valuation& valuation,
		       const class GlobalMarking& marking)
{
  card_t errorplace;
  assert (!myPriority || transition.getPriority () == myPriority);
  assert (valuation.isOK ());
  myPriority = transition.getPriority ();
  if (myEnabled) transition.logEnabled (valuation, myEnabled);
  if (Graph::encode (myStateBuf, marking, *net.getInitMarking (),
		     &errorplace)) {
    mySrcDeadlock = false;
    enum Net::Status status = net.isReject (marking);
    switch (status) {
    case Net::OK:
      break;
    case Net::Fatal:
      flagFatal ();
      // fall through
    case Net::Error:
      break;
    }
    return report (transition, valuation, marking, status != Net::OK);
  }

  reject (transition, valuation, marking, 0, &net.getPlace (errorplace));
  return true;
}

bool
StateReporter::addState (const void* state,
			 size_t size,
			 bool hidden)
{
  if (size > myMaxSize) myMaxSize = size;
  if (size < myMinSize) myMinSize = size;
  if (hidden && myLocal)
    return myLocal->add (state, size), false; // local state suppressed
  if (!mySuccessors && mySuppressed && mySuppressed->size () &&
      mySuppressed->add (state, size))
    return mySuccessors = One, false; // path compression applied
  mySuccessors = Many;
  return do_addState (state, size);
}

word_t*
StateReporter::popState (bool tail, size_t& size)
{
  word_t* buf;
  if (myLocal) {
    if (!(buf = myLocal->pop (tail, size))) {
      if (myLocal->getNumStates ())
	addEvents ();
      myLocal->clear ();
    }
    else
      return buf;
  }
  if (mySuppressed) {
    if (const size_t num = mySuppressed->size ()) {
      const struct States::state& s = (*mySuppressed)[num - 1];
      switch (mySuccessors) {
      case One:
	buf = new word_t[1 + (s.size - 1) / sizeof (word_t)];
	memcpy (buf, s.buf, size = s.size);
	mySuccessors = None;
	return buf;
      case Many:
	do_addState (s.buf, s.size);
	mySuccessors = None;
	// fall through
      case None:
	mySuppressed->clear ();
      }
    }
  }
  if ((buf = do_popState (tail, size)) && mySuppressed)
    if (!mySuppressed->add (buf, size))
      assert (false);
  return buf;
}

void
StateReporter::addEvents ()
{
}

void
StateReporter::reject (const class Transition& transition,
		       const class Valuation& valuation,
		       const class GlobalMarking& marking,
		       const char* reason,
		       const class Place* place)
{
  reject ();
  if (valuation.getError () == errFatal)
    flagFatal ();
  transition.report (valuation, marking, reason, place, printer);
}
